/*
 * $Id: RNodeClientFactoryImpl.java,v 1.11 2005/10/16 15:44:54 rampil Exp $
 * Copyright (c) 2005 LOGICAL-PARADOX.ORG
 */
package org.logical_paradox.rss.router;
import java.net.MalformedURLException;
import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.NoSuchElementException;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.logical_paradox.rss.ModuleException;
import org.logical_paradox.rss.RSSSubModule;
import org.logical_paradox.rss.dsr.RSSDistributedServiceRegistry;
import org.logical_paradox.rss.dsync.SyncCommand;
import org.logical_paradox.rss.dsync.SyncQueue;
import org.logical_paradox.rss.dsync.SyncQueueException;
import org.logical_paradox.rss.http.URLPool;
import org.logical_paradox.rss.lookup.LookupFactory;
import org.logical_paradox.rss.router.algorithm.AllocatingAlgorithm;
import org.logical_paradox.rss.router.algorithm.AllocationException;

/**
 * RNodeClientFactoryImpl()
 * oHm[hNCAg쐬t@Ng<br>
 * KvɉĐVCX^XǉDKvȂꍇ́Ĉ̂Ԃ
 * @author satoshi akabane@logical-paradox.org
 * @version $Revision: 1.11 $
 */
public class RNodeClientFactoryImpl extends UnicastRemoteObject implements RNodeClientFactory, RSSSubModule {
	/** K[ */
	private static final Log log = LogFactory.getLog(RNodeClientFactoryImpl.class);
	/** RtBOt@C */
	public static final String RSS_ROUTING_NODE_CONFIG = "conf/rssroutingfactory.conf";

	private Hashtable searchNodes = null;						// ̃NCAgt@NgǗĂTm[h
	private String factoryServiceName = null;					// ̃T[rX̖O
	private RSSRoutingNodeConfig config = null;				// ̃t@Ng̃RtBO
	private URLPool unTracedURLs = null;						// URL
	private int rnodeClientCapacity = 32;						// Ǌ\ȃNCAg̐
	private SyncQueue syncQueue = new SyncQueue();				// TbNp̃L[
	private SyncQueue rejectedCommandQueue = new SyncQueue();	// ۂĂbNvۊǂL[
	private AllocatingAlgorithm allocater = null;				// zASY
	private String DSRURL;

	/**
	 * RXgN^
	 */
	public RNodeClientFactoryImpl() throws RemoteException {
	}

	/**
	 * 
	 * @param dsrurl DSR
	 * @throws RemoteException Ɏs
	 */
	public void init(String dsrurl) throws RemoteException {
		DSRURL = dsrurl;
		if(DSRURL == null || DSRURL.trim().length() == 0) {
			throw new RemoteException("dsr service name is empty");
		}

		unTracedURLs = new URLPool();
		searchNodes = new Hashtable();

		try {
			// oHm[hIqɃT[rXC
			log.info("UT[rXWXg'" + DSRURL + "'ɃT[rXCĂ܂");

			factoryServiceName = LookupFactory.getLookup(DSRURL).bind(RSSDistributedServiceRegistry.RSS_SERVID_ROUTING_NODE_CLIENT_FACTORY, this);
			log.info("oHm[h: " + getNodeId());

			log.info("T[rXCɐ܂");
		} catch(Exception e) {
			log.info("T[rXCɎs܂");
			throw (RemoteException)e;
		}
	}

	/**
	 * RXgN^
	 */
	public RNodeClientFactoryImpl(String selectorServiceUrl) throws RemoteException {
		init(selectorServiceUrl);
	}

	/**
	 * ̃t@Ng̖OԂ
	 */
	public String getNodeId() throws RemoteException {
		return factoryServiceName;
	}

	/**
	 * Tm[hoCh
	 */
	public RoutingNodeClient bindSearchNode() throws RemoteException {
		log.info("Tm[hoChĂ܂");

		// Tm[hoCh鏈
		RoutingNodeClient client = attachSearchNode();

		try {
			log.info("Tm[hCO'" + client.getNodeId() + "'Ƃēo^܂");
		} catch(Exception e) {
			throw new RemoteException(e.getMessage());
		}
		return client;
	}

	/**
	 * Tm[h؂藣
	 */
	public void unbindSearchNode(String robotId) throws RemoteException {
		log.info("Tm[h'" + robotId + "'؂藣Ă܂");

		boolean exitf = true;
		if(detachSearchNode(robotId)) {
			exitf = false;
		}

		if(exitf) {
			// NǗĂȂTm[h
			log.info("ǗĂȂTm[h'" + robotId + "'؂藣Ƃ܂");
			throw new RemoteException("no routing manager manages the robot '" + robotId + "'.");
		}

		log.info("Tm[h'" + robotId + "'؂藣܂");
	}

	/**
	 * Tm[hoCh
	 */
	public RoutingNodeClient attachSearchNode() throws RemoteException {
		RoutingNodeClientImpl rn = new RoutingNodeClientImpl(this);
		searchNodes.put(rn.getNodeId(), rn);

		// zASYɃm[hǉ
		allocater.addChildNode(rn.getNodeId());

		return rn;
	}

	/**
	 * Tm[h؂藣
	 * @param robotId Tm[hID
	 */
	public boolean detachSearchNode(String robotId) {
		boolean rc = false;
		synchronized(searchNodes) {
			RoutingNodeClient rnc = (RoutingNodeClient)searchNodes.get(robotId);
			if(rnc == null || searchNodes.remove(robotId) == null) {
				// Tm[h̐؂藣Ɏsꍇ
				// ʏ픭ȂCɐ؂藣ĂꍇȂ
				rc = false;
			} else {
				// zASYm[h폜
				allocater.removeChildNode(robotId);
				// Tm[hɃ[XĂ郊[gIuWFNg
				try {
					UnicastRemoteObject.unexportObject(rnc, true);
				} catch (NoSuchObjectException e) {
					// 
				}
				rc = true;
			}
		}

		return rc;
	}

	/**
	 * URLt@Ng̃v[폜
	 */
	public void removeUnTracedURL(String url) throws RemoteException {
		try {
			synchronized(unTracedURLs) {
				unTracedURLs.remove(url);
			}
		} catch(MalformedURLException ue) {
			throw new RemoteException(ue.getMessage());
		}
	}

	/**
	 * t@NgɐVURLǉ
	 */
	public void addUnTracedURL(String url) throws RemoteException {
		try {
			unTracedURLs.add(url);
		} catch(MalformedURLException ue) {
			throw new RemoteException(ue.getMessage());
		}
	}

	/**
	 * Kȉʃm[hɑ΂āCURL蓖Ă
	 */
	public void allocateURL() {
		try {
			synchronized(unTracedURLs) {
				String[] domains = unTracedURLs.domains();
				for(int i = 0; i < domains.length; i++) {
					String domain = domains[i];
					String[] urls = unTracedURLs.urls(domain);

					for(int j = 0; j < urls.length; j++) {
						try {
							String destNodeId = allocater.allocate(urls[j]);
							rejectedCommandQueue.add(new SyncCommand(urls[j], SyncCommand.COMMAND_FORWARD_URL, getNodeId(), destNodeId));
							unTracedURLs.remove(urls[j]);
						} catch(RemoteException re) {
							re.printStackTrace();
						} catch(AllocationException ae) {
							ae.printStackTrace();
						} catch(SyncQueueException se) {
							// ̂Ăꂽ
						} catch(MalformedURLException ue) {
							// ̂Ăꂽ
						}
					}
				}
			}
		} catch(NoSuchElementException nse) {
			// XgɂȂꍇDȂ
		}
	}

	/**
	 * T̋ĂIPAhX͈͂̈ꗗԂD
	 * @return TĂIPAhX͈͂̃Xg
	 * @throws RemoteException ɔO
	 */
	public String[] getAllowedRoutingIPRanges() throws RemoteException {
		return config.getRoutingIPRanges();
	}

	/**
	 * Tm[ḧꗗԂ
	 * @return Tm[h
	 * @throws RemoteException ɔO
	 */
	public String[] getClientNames() throws RemoteException {
		Vector tmp = new Vector();
		synchronized(searchNodes) {
			Enumeration e = searchNodes.keys();
			while(e.hasMoreElements()) {
				tmp.addElement(e.nextElement());
			}
		}

		if(tmp.size() == 0) {
			return null;
		} else {
			return (String[])tmp.toArray(new String[1]);
		}
	}

	/**
	 * w肳ꂽm[hɈĂTbN̍Ԃ
	 * @param rk m[hID
	 * @return Ăɓ͂R}h
	 * @throws RemoteException ɔO
	 */
	public SyncCommand[] getModifiedSyncData(String rk) throws RemoteException {
		allocateURL();				// URLv[ɎcĂ̂CR}h֕ϊĂ݂
		SyncCommand[] o = rejectedCommandQueue.get(rk);
		rejectedCommandQueue.clear(rk);
		return o;
	}

	/**
	 * TbN̍ۑD
	 * @param locks R}h
	 * @throws Exception R}hɔO
	 */
	public void sync(SyncCommand[] locks) throws Exception {
		if(locks == null || locks.length == 0) {
			return;
		}
		log.info(locks.length + "̓f[^M܂");

		Vector receivedCommands = new Vector(Arrays.asList(locks));
		SyncCommand[] rejectedCommands = syncQueue.add(locks);
		receivedCommands.removeAll(Arrays.asList(rejectedCommands));	// MR}h - ۂꂽ = 󗝂ꂽ

		for(int i = 0; i < receivedCommands.size(); i++) {
			SyncCommand command = (SyncCommand)receivedCommands.elementAt(i);
			if(command.getType() == SyncCommand.COMMAND_FORWARD_URL) {
				// URLMꍇCURLv[Ɋi[
				unTracedURLs.add((String)command.getObject());
			} else if(command.getType() == SyncCommand.COMMAND_SHUTDOWN) {
				// Vbg_ER}h̏ꍇĈ܂܎󗝂
				rejectedCommandQueue.add(command);
			}
		}

		// out queueɋۂꂽR}ho^
		rejectedCommandQueue.add(rejectedCommands);
	}

	public static void main(String[] args) {
		try {
			RNodeClientFactoryImpl factory = new RNodeClientFactoryImpl();
			factory.startModule();
		} catch(Exception e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * W[Rg[pC^[tF[X
	 * W[Jn
	 */
	public void startModule() throws ModuleException {
		try {
			RSSRoutingNodeConfig conf = new RSSRoutingNodeConfig(RSS_ROUTING_NODE_CONFIG);

			init(conf.getRSSDSR());
			rnodeClientCapacity = conf.getRSSRNodeClientCapacity();

			// zASY̐ݒ
			allocater = conf.getAllocatingAlgorithm();
			allocater.setRoutingAlgorithm(conf.getRoutingAlgorithm());

			// configɐݒ肳Ă鏉URLĈ̂Ƃēo^
			String defaultURLsLine = conf.getRSSFactoryDefaultURLs();
			if(defaultURLsLine != null) {
				String[] defaultURLs = defaultURLsLine.split(",");
				for(int i = 0; i < defaultURLs.length; i++) {
					addUnTracedURL(defaultURLs[i]);
				}
			}
			config = conf;

			System.err.println("RNodeClientFactory '" + factoryServiceName + "' is started");
			log.info("oHm[hNCAgt@NgT[rX'" + factoryServiceName + "'Jn܂");
		} catch(Exception e) {
			throw new ModuleException(e);
		}
	}

	/**
	 * W[Rg[pC^[tF[X::
	 * W[ꎞ~
	 */
	public void suspendModule() throws ModuleException {
	}

	/**
	 * W[Rg[pC^[tF[X::
	 * ~Ă郂W[̎sĊJ
	 */
	public void continueModule() throws ModuleException {
	}

	/**
	 * W[Rg[pC^[tF[X::
	 * W[~
 	 * ̃\bhĂ΂ꂽꍇCW[͈Sȕ@ŏI
	 * IW[́CVCX^XȂƊJn邱Ƃ͂łȂ
	 */
	public void stopModule() throws ModuleException {
	}

	/**
	 * T[rXԂ
	 * @return T[rX
	 * @throws RemoteException T[rX擾Ɏs
	 */
	public String getName() throws RemoteException {
		return factoryServiceName;
	}
	/**
	 * SCT[rXԂ
	 * @return SCT[rX
	 * @throws RemoteException T[rX擾Ɏs
	 */
	public String getFqsn() throws RemoteException {
		return factoryServiceName + "@" + config.getRSSDSR();
	}
	/**
	 * T[rX~
	 * RMIT[o[ƂĎĂꍇCunexport
	 * @param force true:~ / false:Sɒ~
	 * @throws RemoteException T[rX~Ɏs
	 */
	public void shutdown(boolean force) throws RemoteException {
		log.info("T[o[~܂");

		// STm[hɑ΂āC~R}h𔭍s
		String[] clientNodes = getClientNames();
		if(clientNodes != null && clientNodes.length > 0) {
			log.info("Tm[h̒~҂Ă܂..");
			SyncCommand[] commands = new SyncCommand[clientNodes.length];
			for(int i = 0; i < clientNodes.length; i++) {
				SyncCommand cmd = new SyncCommand();
				cmd.setType(SyncCommand.COMMAND_SHUTDOWN);
				cmd.setTo(clientNodes[i]);
				cmd.setFrom(getNodeId());
				cmd.setObject(clientNodes[i]);
				commands[i] = cmd;
			}
			try {
				sync(commands);
			} catch(RemoteException re) {
				throw re;
			} catch (Exception e) {
				e.printStackTrace();
				throw new RemoteException(e.getMessage());
			}

			log.info("Tm[hSĒ~܂");
		}

		// R}hꍇCTm[hSĒ~̂҂
		try {
			while(searchNodes.size() > 0) {
				Thread.sleep(1000);
			}
		} catch (InterruptedException e) {
			log.warn("Tm[h~҂ɁC\Ȃ荞݂܂F", e);
		}

		// DSR؂藣
		try {
			log.info("oHm[hDSR؂藣Ă܂");
			LookupFactory.getLookup(DSRURL).unbind(RSSDistributedServiceRegistry.RSS_SERVID_ROUTING_NODE_CLIENT_FACTORY, factoryServiceName);
		} catch (Exception e1) {
			throw new RemoteException(e1.getMessage());
		}

		UnicastRemoteObject.unexportObject(this, force);

		log.info("T[o[~܂");
	}
}
